# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for declaring the datatype."""
import base64
from numbers import Number
from typing import (
    Any,
    Dict,
    Union,
)

import numpy as np
import pythonflow as pf

from qctrlcommons.graph import Graph
from qctrlcommons.node.registry import (
    NODE_REGISTRY,
    Node,
)


class DataType:
    """Framework for supporting data types which are not JSON serializable by
    default.

    Attributes
    ----------
    _type: Callable
         data type
    object_key: str
        key name for the data
    """

    _type = None
    object_key = None

    def can_encode(self, obj: Any) -> bool:
        """Checks that the object can be encoded with this class. Default
        behaviour is to check that the object is an instance of _type.

        Parameters
        ----------
        obj : Any
            object to be examined.

        Returns
        -------
        bool
            True if the object can be encoded, False otherwise.
        """
        if self._type is None:
            raise RuntimeError("_type not set for: {}".format(self))

        return isinstance(  # pylint:disable=isinstance-second-argument-not-valid-type
            obj, self._type
        )

    def encode(self, obj) -> dict:
        """Encodes the object. Result should be JSON serializable. To be
        overridden by subclass.

        Parameters
        ----------
        obj : Any
            object to be encoded.
        """
        raise NotImplementedError

    def can_decode(self, obj: dict) -> bool:
        """Checks that the object can be decoded with this class. Default
        behaviour is to check that the object_key exists in the object.

        Parameters
        ----------
        obj : Any
            object to be examined.

        Returns
        -------
        bool
            True if the object can be decoded, False otherwise.
        """
        if self.object_key is None:
            raise RuntimeError("object_key not set for: {}".format(self))

        return self.object_key in obj

    def decode(self, obj: dict):
        """Decodes the object. To be overridden by subclass.

        Parameters
        ----------
        obj: dict
            object to be decoded.
        """
        raise NotImplementedError


class NumpyScalar(DataType):  # pylint:disable=abstract-method
    """Handle np.number serialization."""

    _type = np.number

    def encode(self, obj: np.number) -> Any:
        """cast to python builtin int or float"""
        return obj.item()

    def can_decode(self, obj: dict) -> bool:
        return False


class NumpyArray(DataType):
    """Represent NumpyArray Model."""

    _type = np.ndarray
    object_key = "base64_encoded_array"

    def encode(self, obj):
        if obj.flags["C_CONTIGUOUS"]:
            obj_data = obj.data
        else:
            cont_obj = np.ascontiguousarray(obj)
            assert cont_obj.flags[  # pylint:disable=unsubscriptable-object
                "C_CONTIGUOUS"
            ]
            obj_data = cont_obj.data

        data_b64 = base64.b64encode(obj_data)
        return {
            self.object_key: data_b64.decode("ascii"),
            "dtype": str(obj.dtype),
            "shape": list(obj.shape),
        }

    def decode(self, obj):
        return np.frombuffer(
            base64.b64decode(obj[self.object_key]), dtype=obj["dtype"]
        ).reshape(obj["shape"])


class NumpyComplexNumber(DataType):
    """Represent NumpyComplexNumber Model."""

    _type = np.complexfloating
    object_key = "base64_encoded_data"

    def encode(self, obj) -> dict:
        obj_data = obj.data
        data_b64 = base64.b64encode(obj_data)
        return {
            self.object_key: data_b64.decode("utf-8"),
            "dtype": str(obj.dtype),
            "shape": list(obj.shape),
        }

    def decode(self, obj: dict):
        cast_to = getattr(np, obj["dtype"])
        return cast_to(
            np.frombuffer(
                base64.b64decode(obj[self.object_key]), dtype=obj["dtype"]
            ).reshape(obj["shape"])
        )


class ComplexNumber(DataType):
    """Represents ComplexNumber Model."""

    _type = complex
    object_key = "encoded_complex"

    def encode(self, obj) -> dict:
        return {self.object_key: True, "real": obj.real, "imag": obj.imag}

    def decode(self, obj: dict):
        return complex(obj["real"], obj["imag"])


class NodeDataType(DataType):
    """Represents NodeDataType for pythonflow operation."""

    _type = pf.core.func_op
    object_key = "pythonflow_op"

    def encode(
        self, obj: Union[pf.core.func_op, Number, np.complexfloating, np.ndarray, Node]
    ) -> dict:
        """Encode object to dictionary.

        Parameters
        ----------
        obj: Union[pf.core.func_op, Number, np.complexfloating, np.ndarray, Node]
            object to be encoded.

        Returns
        -------
        dict
            serialized node
        """
        if isinstance(obj, Node):
            obj_id = obj.node_id
            operation_name = obj.name
            args = obj._args  # pylint:disable=protected-access
            kwargs = obj._kwargs  # pylint:disable=protected-access
        else:
            obj_id = obj.name
            if isinstance(obj.target, Node):
                operation_name = obj.target.name
            else:
                operation_name = (
                    obj.target.__name__ if hasattr(obj.target, "__name__") else obj.name
                )
            args = obj.args
            kwargs = obj.kwargs

        return {
            self.object_key: True,
            "id": obj_id,
            "operation_name": operation_name,
            "args": args,
            "kwargs": kwargs,
        }

    def decode(self, obj: Union[Dict, Number, np.complexfloating, np.ndarray]) -> Node:
        """Decode to Node object.

        Parameters
        ----------
        obj: Union[Dict, Number, np.complexfloating, np.ndarray]
            object to be decoded.

        Returns
        -------
        Mode
            Custom class to represent the node in dataflow graph.
        """
        node_cls = NODE_REGISTRY.get_node_cls(obj["operation_name"])
        return node_cls(obj["id"], obj["args"], obj["kwargs"])

    def can_encode(self, obj) -> bool:
        return isinstance(obj, (pf.core.func_op, Node))


class GraphDataType(DataType):
    """Represent GraphDataType for pythonflow Graph."""

    _type = pf.core.Graph
    object_key = "pythonflow_graph"

    def encode(self, obj: Union[pf.Graph, Graph]) -> dict:
        """Convert pythonflow graph to dict, all the operations from graph will
        be encoded with NodeDataType.

        Parameters
        ----------
        obj: Union[pf.Graph, Graph]
            object to be encoded.

        Returns
        -------
        dict
            serialized graph.
        """
        return {
            self.object_key: True,
            "operations": obj.operations,
            "dependencies": obj.dependencies,
        }

    def decode(self, obj: dict) -> Graph:
        """Decode to pythonflow graph.

        Parameters
        ----------
        obj: dict
            object to be decoded.

        Returns
        -------
        Graph:
            a custom structure for encoding and decoding pythonflow graph.
        """
        if not obj.get("operations", None):
            raise KeyError("Missing operations. It cannot be encoded to graph")

        return Graph(obj["operations"], obj["dependencies"])

    def can_encode(self, obj) -> bool:
        return isinstance(obj, (pf.core.Graph, Graph))
