# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for handling exceptions."""
import json
from functools import wraps
from typing import (
    Callable,
    List,
    Union,
)


class QctrlException(Exception):
    """Base class for all qctrl-related errors."""


class QctrlGqlException(QctrlException):
    """Exception for GraphQL errors.

    Attributes
    ----------
    message: Union[List, str]
        exception messages.
    **kwargs
        Arbitrary keyword argument.

    """

    def __init__(self, message: Union[List, str] = None, **kwargs):
        super().__init__(message, **kwargs)
        self._errors = []

        if message is None:
            pass

        elif isinstance(message, list):
            self._errors = message.copy()

        else:

            # error message might be a list of errors
            try:
                errors = json.loads(message)
                assert isinstance(errors, list)

            # otherwise single error message
            except (json.JSONDecodeError, AssertionError):
                self._errors.append(message)

            else:
                self._errors = errors

    def __str__(self):
        if not self._errors:
            return "An error occurred while executing the query"

        if len(self._errors) == 1:
            return self._errors[0]

        lines = ["The following errors occurred:"]

        for error in self._errors:
            lines.append(f"- {error}")

        return "\n".join(lines)


def wrap_with_qctrl_exception(func: Callable) -> Callable:
    """Wraps any occuring exception with a `QctrlException`. The original
    exception can be accessed through `QctrlException.__cause__`

    Parameters
    ----------
    func : Callable
        a callable object.

    Returns
    -------
    Callable:
        the result of the callable.
    """

    @wraps(func)
    def exception_wrapped_f(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as exc:
            raise QctrlException(f"{exc.__str__()}") from exc

    return exception_wrapped_f
