# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for CoreNode."""
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Union,
)

import forge
import numpy as np

from qctrlcommons.node.module import ModuleNode

from . import types


class CoreNode(ModuleNode):
    """Abstract class to represent the module is qctrlcore."""

    _module_name = "qctrlcore"


class TensorPwc(CoreNode):
    r"""
    Creates a piecewise-constant function of time.

    Parameters
    ----------
    durations : np.ndarray (1D, real)
        The durations :math:`\{\delta t_n\}` of the constant segments.
    values: Tensor
        The values :math:`\{v_n\}` of the function on the constant segments.
        The first dimension must be the same length as `durations`.

    Returns
    -------
    TensorPwc
        The piecewise-constant function of time :math:`v(t)`, satisfying
        :math:`v(t)=v_n` for :math:`t_{n-1}\leq t\leq t_n`, where
        :math:`t_0=0` and :math:`t_n=t_{n-1}+\delta t_n`.
    """

    name = "tensor_pwc"
    _module_attr = "TensorPwc"
    args = [
        forge.arg("durations", type=np.ndarray),
        forge.arg("values", type=types.Tensor),
    ]

    rtype = types.TensorPwc


class PwcSignal(CoreNode):
    r"""
    Creates a piecewise-constant signal (scalar-valued function of time).

    Use this function to create a piecewise-constant signal in which the
    constant segments all have the same duration.

    Parameters
    ----------
    values : Tensor(1D)
        The values :math:`\{\alpha_n\}` of the constant segments.
    duration : float
        The total duration :math:`\tau` of the signal.

    Returns
    -------
    TensorPwc(1D)
        The piecewise-constant function of time :math:`\alpha(t)`, satisfying
        :math:`\alpha(t)=\alpha_n` for :math:`t_{n-1}\leq t\leq t_n`, where
        :math:`t_n=n\tau/N` (where :math:`N` is the number of values
        in :math:`\{\alpha_n\}`).

    See Also
    --------
    tensor_pwc : Corresponding function with support for segments of different
        durations.
    """

    name = "pwc_signal"
    _module_attr = "create_pwc_signal"
    args = [forge.arg("values", type=types.Tensor), forge.arg("duration", type=float)]
    rtype = types.TensorPwc


class ComplexPwcSignal(CoreNode):
    r"""
    Creates a complex piecewise-constant signal from moduli and phases.

    Use this function to create a complex piecewise-constant signal from
    moduli and phases defined for each segment, in which the constant segments
    all have the same duration.

    Parameters
    ----------
    moduli : Tensor(1D, real)
        The moduli :math:`\{\Omega_n\}` of the constant segment values.
    phases : Tensor(1D, real)
        The phases :math:`\{\phi_n\}` of the constant segment values. Must
        have the same length as `moduli`.
    duration : float
        The total duration :math:`\tau` of the signal.

    Returns
    -------
    TensorPwc(1D, complex)
        The piecewise-constant function of time :math:`v(t)`, satisfying
        :math:`v(t)=\Omega_n e^{i\phi_n}` for :math:`t_{n-1}\leq t\leq t_n`,
        where :math:`t_n=n\tau/N` (where :math:`N` is the number of
        values in :math:`\{\Omega_n\}` and :math:`\{\phi_n\}`).

    See Also
    --------
    pwc_signal : Corresponding function for creating signals from
        (possibly-complex) values.
    """

    name = "complex_pwc_signal"
    _module_attr = "create_complex_pwc_signal"
    args = [
        forge.arg("moduli", type=types.Tensor),
        forge.arg("phases", type=types.Tensor),
        forge.arg("duration", type=float),
    ]
    rtype = types.TensorPwc


class PwcOperatorHermitianPart(CoreNode):
    r"""
    Creates the Hermitian part of a piecewise-constant operator.

    Parameters
    ----------
    operator : TensorPwc(3D)
        The operator :math:`A(t)`.

    Returns
    -------
    TensorPwc(3D)
        The Hermitian part :math:`\frac{1}{2}(A(t)+A^\dagger(t))`.
    """

    name = "pwc_operator_hermitian_part"
    _module_attr = "get_pwc_operator_hermitian_part"
    args = [
        forge.arg("operator", type=types.TensorPwc),
    ]
    rtype = types.TensorPwc


class PwcSum(CoreNode):
    r"""
    Creates the sum of multiple piecewise-constant terms.

    Parameters
    ----------
    terms : list[TensorPwc]
        The individual piecewise-constant terms :math:`\{v_j(t)\}` to sum. All
        terms must have the same duration and values of the same shape, but
        may have different segmentations (different numbers of segments of
        different durations).

    Returns
    -------
    TensorPwc
        The piecewise-constant function of time :math:`\sum_j v_j(t)`.
    """

    name = "pwc_sum"
    _module_attr = "get_pwc_sum"
    args = [forge.arg("terms", type=List[types.TensorPwc])]
    rtype = types.TensorPwc


class TargetOperator(CoreNode):
    r"""
    Creates information about the target for system time evolution.

    Nodes created with this function contain two types of information: the
    target gate for the system time evolution, and the projection operator
    that defines the subspace of interest for robustness.

    Parameters
    ----------
    operator : np.ndarray
         The target gate :math:`U_\mathrm{target}`. Must be a non-zero partial
         isometry.
    filter_function_projector : np.ndarray, optional
        The orthogonal projection matrix :math:`P` onto the subspace used for
        filter function calculations. If you provide a value then it must be
        Hermitian and idempotent. Defaults to the identity matrix.

    Returns
    -------
    TargetOperator
        The node containing the specified target information.

    See Also
    --------
    infidelity_pwc, infidelity_stf

    Notes
    -----
    The target gate :math:`U_\mathrm{target}` is a non-zero partial isometry,
    which means that it can be expressed in the form
    :math:`\sum_j \left|\psi_j\right>\left<\phi_j\right|`, where
    :math:`\left\{\left|\psi_j\right>\right\}` and
    :math:`\left\{\left|\phi_j\right>\right\}` both form (non-empty)
    orthonormal, but not necessarily complete, sets. Such a target represents
    a target state :math:`\left|\psi_j\right>` for each initial state
    :math:`\left|\phi_j\right>`. The resulting operational infidelity is 0
    if and only if, up to global phase, each initial state
    :math:`\left|\phi_j\right>` maps exactly to the corresponding final state
    :math:`\left|\psi_j\right>`.

    The filter function projector :math:`P` is an orthogonal projection
    matrix, which means that it satisfies :math:`P=P^\dagger=P^2`. The image
    of :math:`P` defines the set of initial states from which the calculated
    filter function measures robustness.
    """

    name = "target"
    _module_attr = "Target"
    args = [
        forge.arg("operator", type=np.ndarray),
        forge.arg("filter_function_projector", type=Optional[np.ndarray], default=None),
    ]
    rtype = types.TargetOperator


class RealFourierPwcSignal(CoreNode):
    r"""
    Creates a piecewise-constant signal constructed from Fourier components.

    Use this function to create a signal defined in terms of Fourier
    (sine/cosine) basis signals that can be optimized by varying their
    coefficients and, optionally, their frequencies.

    This function supports three different modes: fixed frequencies (you
    directly provide the frequencies for the basis signals), optimizable
    frequencies (you provide the number of frequencies, and the frequency
    values can be tuned by the optimizer), and randomized frequencies (you
    provide the number of frequencies, and the frequency values are chosen at
    random at the start of each optimization and then held constant).

    Parameters
    ----------
    duration : float
        The total duration :math:`\tau` of the signal.
    segments_count : int
        The number of segments :math:`N` to use for the piecewise-constant
        approximation.
    initial_coefficient_lower_bound : float, optional
        The lower bound :math:`c_\mathrm{min}` on the initial coefficient
        values. Defaults to -1.
    initial_coefficient_upper_bound : float, optional
        The upper bound :math:`c_\mathrm{max}` on the initial coefficient
        values. Defaults to 1.
    fixed_frequencies : list[float], optional
        The fixed non-zero frequencies :math:`\{f_m\}` to use for the Fourier
        basis. Must be non-empty if provided. Must be specified in the inverse
        units of `duration` (for example if `duration` is in seconds, these
        values must be given in Hertz).
    optimizable_frequencies_count : int, optional
        The number of non-zero frequencies :math:`M` to use, if the
        frequencies can be optimized. Must be greater than zero if provided.
    randomized_frequencies_count : int, optional
        The number of non-zero frequencies :math:`M` to use, if the
        frequencies are to be randomized but fixed. Must be greater than zero
        if provided.

    Returns
    -------
    TensorPwc(1D, real)
        The optimizable, real-valued, piecewise-constant signal built from the
        appropriate Fourier components.

    Notes
    -----
    This function sets the basis signal frequencies :math:`\{f_m\}`
    depending on the chosen mode:

    * For fixed frequencies, you provide the frequencies directly.
    * For optimizable frequencies, you provide the number of frequencies
      :math:`M`, and this function creates :math:`M` unbounded optimizable
      variables :math:`\{f_m\}`, with initial values in the ranges
      :math:`\{[(m-1)/\tau, m/\tau]\}`.
    * For randomized frequencies, you provide the number of frequencies
      :math:`M`, and this function creates :math:`M` randomized constants
      :math:`\{f_m\}` in the ranges :math:`\{[(m-1)/\tau, m/\tau]\}`.

    After this function creates the :math:`M` frequencies :math:`\{f_m\}`, it
    produces the signal

    .. math::
        \alpha^\prime(t) = v_0 +
        \sum_{m=1}^M [ v_m \cos(2\pi t f_m) + w_m \sin(2\pi t f_m) ],

    where :math:`\{v_m,w_m\}` are (unbounded) optimizable variables, with
    initial values bounded by :math:`c_\mathrm{min}` and
    :math:`c_\mathrm{max}`. This function produces the final
    piecewise-constant signal :math:`\alpha(t)` by sampling
    :math:`\alpha^\prime(t)` at :math:`N` equally spaced points along the
    duration :math:`\tau`, and using those sampled values as the constant
    segment values.

    You can use the signals created by this function for chopped random basis
    (CRAB) optimization [1]_.

    References
    ----------
    .. [1] `P. Doria, T. Calarco, and S. Montangero, Physical Review Letters 106, 190501 (2011).
           <https://doi.org/10.1103/PhysRevLett.106.190501>`_
    """

    name = "real_fourier_pwc_signal"
    _module_attr = "create_real_fourier_pwc_signal"
    args = [
        forge.arg("duration", type=float),
        forge.arg("segments_count", type=int),
        forge.arg(
            "initial_coefficient_lower_bound", type=Optional[float], default=None
        ),
        forge.arg(
            "initial_coefficient_upper_bound", type=Optional[float], default=None
        ),
        forge.arg("fixed_frequencies", type=Optional[List[float]], default=None),
        forge.arg("optimizable_frequencies_count", type=Optional[int], default=None),
        forge.arg("randomized_frequencies_count", type=Optional[int], default=None),
    ]
    rtype = types.TensorPwc

    def _evaluate_node(
        self, execution_context: "ExecutionContext", args, kwargs
    ) -> Any:
        """Add variable factory to args.

        Parameters
        ----------
        execution_context : ExecutionContext
            helper class for evaluating the value of the node.
        args
            argument list for the node.
        kwargs
            keyword arguments for the node.

        Returns
        -------
        Any
            the value of the node.
        """
        variable_factory = execution_context.variable_factory
        args.insert(0, variable_factory)
        return super()._evaluate_node(execution_context, args, kwargs)


class SincIntegralFunction(CoreNode):
    r"""
    Creates a function that computes the integral of the sinc function.

    Use this function to create a filter kernel that eliminates frequencies
    that are above a certain cut-off.

    Parameters
    ----------
    cut_off_frequency : float or Tensor
        Upper limit :math:`\omega_c` of the range of frequencies that you want
        to preserve. The filter eliminates components of the signal that have
        higher frequencies.

    Returns
    -------
    Function
        A node representing a function that computes the integral of the sinc
        function.

    Notes
    -----
    The range of frequencies that this kernel lets pass is
    :math:`[-\omega_c, \omega_c]`. After a Fourier transform to convert from
    frequency domain to time domain, this becomes:

    .. math::
        \frac{1}{2\pi} \int_{-\omega_c}^{\omega_c} \mathrm{d}\omega
        e^{i \omega t} = \frac{\sin(\omega_c t)}{\pi t}.

    The function on the right side of the equation is the sinc function.
    Its integral is the sine integral function (Si).
    """

    name = "sinc_integral_function"
    _module_attr = "create_sinc_integral_function"
    args = [
        forge.arg("cut_off_frequency", type=Union[float, types.Tensor]),
    ]
    rtype = types.Function


class GaussianIntegralFunction(CoreNode):
    r"""
    Creates a function that computes the integral of a normalized Gaussian.

    Use this function to create a filter kernel that has a Gaussian shape. A
    Gaussian kernel lets pass frequencies in the range roughly determined
    by its width, and progressively suppresses components outside that
    range.

    Parameters
    ----------
    std : float or Tensor
        Standard deviation :math:`\sigma` of the Gaussian in the time domain.
        The standard deviation in the frequency domain is its inverse, so that
        a high value of this parameter lets fewer frequencies pass.
    offset : float or Tensor, optional
        Center :math:`\mu` of the Gaussian distribution in the time domain.
        Use this to offset the signal in time. Defaults to 0.

    Returns
    -------
    Function
        A node representing a function that computes the integral of the
        Gaussian function.

    Notes
    -----
    The Gaussian that this function integrates is normalized in the time
    domain:

    .. math::
        \frac{e^{-(t-\mu)^2/(2\sigma^2)}}{\sqrt{2\pi\sigma^2}}.

    In the frequency domain, this Gaussian has standard deviation
    :math:`\omega_c= \sigma^{-1}`. The filter it represents therefore
    passes frequencies roughly in the range :math:`[-\omega_c, \omega_c]`.
    """

    name = "gaussian_integral_function"
    _module_attr = "create_gaussian_integral_function"
    args = [
        forge.arg("std", type=Union[float, types.Tensor]),
        forge.arg("offset", type=Optional[Union[float, types.Tensor]]),
    ]
    rtype = types.Function


class ConvolvePwc(CoreNode):
    r"""
    Creates the convolution of a piecewise-constant function with a kernel.

    Parameters
    ----------
    pwc : TensorPwc(1D)
        The piecewise-constant function :math:`\alpha(t)` to convolve.
    kernel_integral : Function
        The node representing the function that computes the integral of the
        kernel :math:`K(t)`.

    Returns
    -------
    Stf(1D)
        The smooth signal :math:`(\alpha * K)(t)`.

    Notes
    -----
    The convolution is

    .. math::
        (\alpha * K)(t) \equiv
        \int_{-\infty}^\infty \alpha(\tau) K(t-\tau) d\tau.

    Convolution in the time domain is equivalent to multiplication in the
    frequency domain, so this function can be viewed as applying a linear
    time-invariant filter (specified via its time domain kernel :math:`K(t)`)
    to :math:`\alpha(t)`.
    """

    name = "convolve_pwc"
    _module_attr = "convolve_pwc"
    args = [
        forge.arg("pwc", type=types.TensorPwc),
        forge.arg("kernel_integral", type=Callable),
    ]
    rtype = types.Stf


class StfSum(CoreNode):
    r"""
    Creates the sum of multiple smooth terms.

    Parameters
    ----------
    terms : list[Stf]
        The individual smooth terms :math:`\{v_j(t)\}` to sum.

    Returns
    -------
    Stf
        The smooth function of time :math:`\sum_j v_j(t)`.
    """

    name = "stf_sum"
    _module_attr = "get_stf_sum"
    args = [forge.arg("terms", type=List[types.Stf])]
    rtype = types.Stf


class SymmetrizePwc(CoreNode):
    r"""
    Creates the symmetrization of a piecewise-constant function.

    Parameters
    ----------
    pwc : TensorPwc
        The piecewise-constant function :math:`v(t)` to symmetrize.

    Returns
    -------
    TensorPwc
        The piecewise-constant function :math:`w(t)` defined by
        :math:`w(t)=v(t)` for :math:`0\leq t\leq \tau` and
        :math:`w(t)=v(2t-\tau)` for :math:`\tau\leq t\leq 2\tau`, where
        :math:`\tau` is the duration of :math:`v(t)`.
    """

    name = "symmetrize_pwc"
    _module_attr = "symmetrize_pwc"
    args = [forge.arg("pwc", type=types.TensorPwc)]
    rtype = types.TensorPwc


class AnchoredDifferenceBoundedVariables(CoreNode):
    r"""
    Creates a sequence of variables with an anchored difference bound.

    Use this function to create a sequence of optimization variables that have
    a difference bound (each variable is constrained to be within a given
    distance of the adjacent variables) and are anchored to zero at the start
    and end (the initial and final variables are within a given distance of
    zero).

    Parameters
    ----------
    count : int
        The number :math:`N` of individual real-valued variables to create.
    lower_bound : float
        The lower bound :math:`v_\mathrm{min}` on the variables.
        The same lower bound applies to all `count` individual variables.
    upper_bound : float
        The upper bound :math:`v_\mathrm{max}` on the variables.
        The same upper bound applies to all `count` individual variables.
    difference_bound : float
        The difference bound :math:`\delta` to enforce between adjacent
        variables.

    Returns
    -------
    Tensor
        The sequence :math:`\{v_n\}` of :math:`N` anchored difference-bounded
        optimization variables, satisfying
        :math:`v_\mathrm{min}\leq v_n\leq v_\mathrm{max}`,
        :math:`|v_{n-1}-v_n|\leq\delta` for :math:`2\leq n\leq N`,
        :math:`|v_1|\leq\delta`, and :math:`|v_N|\leq\delta`.

    See Also
    --------
    bounded_optimization_variable, unbounded_optimization_variable
    """

    name = "anchored_difference_bounded_variables"
    _module_attr = "create_anchored_difference_bounded_variables"
    args = [
        forge.arg("count", type=int),
        forge.arg("lower_bound", type=float),
        forge.arg("upper_bound", type=float),
        forge.arg("difference_bound", type=float),
    ]
    rtype = types.Tensor

    def _evaluate_node(
        self, execution_context: "ExecutionContext", args, kwargs
    ) -> Any:
        """Add variable factory to args.

        Parameters
        ----------
        execution_context : ExecutionContext
            helper class for evaluating the value of the node.
        args
            argument list for the node.
        kwargs
            keyword arguments for the node.

        Returns
        -------
        Any
            the value of the node.
        """
        variable_factory = execution_context.variable_factory
        args.insert(0, variable_factory)
        return super()._evaluate_node(execution_context, args, kwargs)


class DiscretizeStf(CoreNode):
    r"""
    Creates a piecewise-constant function by discretizing a smooth function.

    Use this function to create a piecewise-constant approximation to a smooth
    function (obtained, for example, by filtering an initial
    piecewise-constant function).

    Parameters
    ----------
    stf : Stf
        The smooth function :math:`v(t)` to discretize.
    duration : float
        The duration :math:`\tau` over which discretization should be
        performed. The resulting piecewise-constant function has this
        duration.
    segments_count : int
        The number of segments :math:`N` in the resulting piecewise-constant
        function.
    samples_per_segment : int, optional
        The number of samples :math:`M` of the smooth function to take when
        calculating the value of each segment in the discretization. Defaults
        to 1.

    Returns
    -------
    TensorPwc
        The piecewise-constant function :math:`w(t)` obtained by discretizing
        the smooth function.

    Notes
    -----
    The resulting function :math:`w(t)` is piecewise-constant with :math:`N`
    segments, meaning it has segment values :math:`\{w_n\}` such that
    :math:`w(t)=w_n` for :math:`t_{n-1}\leq t\leq t_n`, where :math:`t_n= n \tau/N`.

    Each segment value :math:`w_n` is the average of samples of :math:`v(t)`
    at the midpoints of :math:`M` equally sized subsegments between
    :math:`t_{n-1}` and :math:`t_n`:

    .. math::
        w_n = \frac{1}{M}
        \sum_{m=1}^M v\left(t_{n-1} + \left(m-\tfrac{1}{2}\right) \frac{\tau}{MN} \right).
    """

    name = "discretize_stf"
    _module_attr = "discretize_stf"
    args = [
        forge.arg("stf", type=types.Stf),
        forge.arg("duration", type=float),
        forge.arg("segments_count", type=int),
        forge.arg("samples_per_segment", type=Optional[int], default=None),
    ]
    rtype = types.Tensor


class HessianMatrix(CoreNode):
    r"""
    Calculates a single Hessian matrix for all the variables.

    The Hessian is a matrix containing all the second partial derivatives
    of the `tensor` with respect to the `variables`.

    Parameters
    ----------
    tensor: Tensor(scalar, real)
        The real scalar tensor :math:`T` whose Hessian matrix you want to
        calculate.

    variables: List[Tensor(real)]
        The list of real variables :math:`\{\theta_i\}` with respect to
        which you want to take the second partial derivatives of the
        tensor. If any of the tensors of the list is not scalar, this
        function treats each of the elements of the tensor as a different
        variable. It does this by flattening all tensors and concatenating
        them in the same sequence that you provided in this list.

    Returns
    -------
    Tensor(2D, real)
        The real Hessian matrix :math:`H` containing the second partial
        derivatives of the `tensor` :math:`T` with respect to the
        `variables` :math:`\{\theta_i\}`.

    Warnings
    --------
    This function currently doesn't support calculating a Hessian matrix for
    a graph which includes a `infidelity_pwc` node if it involves a Hamiltonian
    with degenerate eigenvalues at any segment. In that case, the function
    returns a `nan` Hessian matrix.

    Notes
    -----
    The :math:`(i,j)` element of the Hessian contains the partial
    derivative of the `tensor` with respect to the ith and the jth
    `variables`:

    .. math::
        H_{i,j} = \frac{\partial^2 T}{\partial \theta_i \partial \theta_j}.

    The variables :math:`\{\theta_i\}` follow the same sequence as the
    input list of `variables`. If some of the `variables` are not scalars,
    this function flattens them and concatenates them in the same order of
    the list of `variables` that you provided to create the sequence of
    scalar variables :math:`\{\theta_i\}`.

    If you choose a negative log-likelihood as the tensor :math:`T`, you
    can use this Hessian as an estimate of the Fisher information matrix.
    """
    name = "hessian_matrix"
    _module_attr = "hessian_matrix"
    args = [
        forge.arg("tensor", type=types.Tensor),
        forge.arg("variables", type=List[types.Tensor]),
    ]
    rtype = types.Tensor
