# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for ModuleNode."""
from typing import Callable

from qctrlcommons.node.base import Node


class ModuleNode(Node):
    """Represent ModuleNode class."""

    _module_name = None
    _module_attr = None

    def _get_module(self, execution_context):
        module = getattr(execution_context, self._module_name, None)

        if module is None:
            raise RuntimeError(
                f"module missing from execution_context: {self._module_name}"
            )

        return module

    def _evaluate_node(self, execution_context: "ExecutionContext", args, kwargs):
        """Evaluate the node by executing its callable module attribute.

        Parameters
        ----------
        execution_context : ExecutionContext
            helper class for evaluating the value of the node.
        args
            argument list for the node.
        kwargs
            keyword arguments for the node.

        Returns
        -------
        Any
            the value of the node.
        """
        value: callable = self._get_attr_from_module(
            execution_context, self._module_attr
        )
        return value(*args, **kwargs)

    def _get_attr_from_module(
        self, execution_context: "ExecutionContext", module_attr: str
    ) -> Callable:
        """Get attribute from the execution context.

        Parameters
        ----------
        execution_context : ExecutionContext
            helper class for evaluating the value of node.
        module_attr: str
            execution module (i.e qctrlcore, tensorflow)

        Returns
        -------
        Callable
            method or object in the module

        Raises
        ------
        RuntimeError
            when it cannot be found in the module.
        """
        value = self._get_module(execution_context)
        module_attr_names = module_attr.split(".")

        for name in module_attr_names:
            value = getattr(value, name, None)
            if value is None:
                raise RuntimeError(f"attr missing from module: {module_attr}")

        return value
