# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for Operator Node."""
from typing import (
    Dict,
    List,
)

import forge
import numpy as np

from qctrlcommons.node.module import ModuleNode

from . import types


class OperatorNode(ModuleNode):
    """Represents CoreNode class."""

    _module_name = "qctrlcore"

    def _evaluate_node(
        self, execution_context: "ExecutionContext", args: List, kwargs: Dict
    ):
        """Create the Operator object and then evaluate the node.

        Parameters
        ----------
        execution_context : ExecutionContext
            helper class for evaluating the value of the node.
        args: List
            argument list for the node.
        kwargs: Dict
            keyword arguments for the node.

        Returns
        -------
        Any
            the value of the node.
        """
        # create Operator object
        if kwargs.get("operator", None) is not None:
            kwargs["operator"] = self._get_attr_from_module(
                execution_context, "Operator"
            )(kwargs["operator"])
        else:
            args_list = [arg.name for arg in self.args]
            operator_index = args_list.index("operator")
            args[operator_index] = self._get_attr_from_module(
                execution_context, "Operator"
            )(args[operator_index])
        return super()._evaluate_node(execution_context, args, kwargs)


class PwcOperator(OperatorNode):
    """
    Creates a constant operator multiplied by a piecewise-constant signal.

    Parameters
    ----------
    signal : TensorPwc(1D)
        The piecewise-constant signal :math:`a(t)`.
    operator : np.ndarray
        The operator :math:`A`.

    Returns
    -------
    TensorPwc(3D)
        The piecewise-constant operator :math:`a(t)A`.
    """

    name = "pwc_operator"
    _module_attr = "create_pwc_operator"
    args = [
        forge.arg("signal", type=types.TensorPwc),
        forge.arg("operator", type=np.ndarray),
    ]
    rtype = types.TensorPwc


class ConstantPwcOperator(OperatorNode):
    r"""
    Creates a constant piecewise-constant operator over a specified duration.

    Parameters
    ----------
    duration : float
        The duration :math:`\tau` for the resulting piecewise-constant
        operator.
    operator : np.ndarray
        The operator :math:`A`.

    Returns
    -------
    TensorPwc(3D)
        The constant operator :math:`t\mapsto A` (for :math:`0\leq t\leq\tau`).
    """

    name = "constant_pwc_operator"
    _module_attr = "create_constant_pwc_operator"
    args = [forge.arg("duration", type=float), forge.arg("operator", type=np.ndarray)]
    rtype = types.TensorPwc


class StfOperator(OperatorNode):
    """
    Creates a constant operator multiplied by a smooth signal.

    Parameters
    ----------
    signal : Stf(1D)
        The smooth signal :math:`a(t)`.
    operator : np.ndarray
        The operator :math:`A`.

    Returns
    -------
    Stf(3D)
        The smooth operator :math:`a(t)A`.
    """

    name = "stf_operator"
    _module_attr = "create_stf_operator"
    args = [forge.arg("signal", type=types.Stf), forge.arg("operator", type=np.ndarray)]
    rtype = types.Stf


class ConstantStfOperator(OperatorNode):
    r"""
    Creates a constant smooth operator.

    Parameters
    ----------
    operator : np.ndarray
        The operator :math:`A`.

    Returns
    -------
    Stf(3D)
        The smooth operator :math:`t\mapsto A`.
    """

    name = "constant_stf_operator"
    _module_attr = "create_constant_stf_operator"
    args = [forge.arg("operator", type=np.ndarray)]
    rtype = types.Stf
